@@ -6,6 +6,7 @@
 #include "aco_ir.h"
 
 #include "util/bitscan.h"
+#include "util/bitset.h"
 #include "util/macros.h"
 
 #include <limits>
@@ -40,19 +41,17 @@ struct VOPDInfo {
 
 struct InstrInfo {
    Instruction* instr;
-   int32_t priority;
+   int16_t wait_cycles;          /* estimated remaining cycles until instruction can be issued. */
    mask_t dependency_mask;       /* bitmask of nodes which have to be scheduled before this node. */
+   mask_t write_for_read_mask;   /* bitmask of nodes in the DAG that have a RaW dependency. */
    uint8_t next_non_reorderable; /* index of next non-reorderable instruction node after this one. */
-   bool potential_clause; /* indicates that this instruction is not (yet) immediately followed by a
-                             reorderable instruction. */
 };
 
 struct RegisterInfo {
    mask_t read_mask; /* bitmask of nodes which have to be scheduled before the next write. */
-   int8_t latency;   /* estimated latency of last register write. */
-   uint8_t direct_dependency : 4;     /* node that has to be scheduled before any other access. */
-   uint8_t has_direct_dependency : 1; /* whether there is an unscheduled direct dependency. */
-   uint8_t padding : 3;
+   uint16_t latency : 11; /* estimated outstanding latency of last register write outside the DAG. */
+   uint16_t direct_dependency : 4;     /* node that has to be scheduled before any other access. */
+   uint16_t has_direct_dependency : 1; /* whether there is an unscheduled direct dependency. */
 };
 
 struct SchedILPContext {
@@ -60,10 +59,13 @@ struct SchedILPContext {
    bool is_vopd = false;
    InstrInfo nodes[num_nodes];
    RegisterInfo regs[512];
+   BITSET_DECLARE(reg_has_latency, 512) = { 0 };
    mask_t non_reorder_mask = 0; /* bitmask of instruction nodes which should not be reordered. */
    mask_t active_mask = 0;      /* bitmask of valid instruction nodes. */
    uint8_t next_non_reorderable = UINT8_MAX; /* index of next node which should not be reordered. */
    uint8_t last_non_reorderable = UINT8_MAX; /* index of last node which should not be reordered. */
+   bool potential_partial_clause; /* indicates that last_non_reorderable is the last instruction in
+                                     the DAG, meaning the clause might continue outside of it. */
 
    /* VOPD scheduler: */
    VOPDInfo vopd[num_nodes];
@@ -80,7 +82,7 @@ struct SchedILPContext {
 bool
 can_reorder(const Instruction* const instr)
 {
-   if (instr->isVALU())
+   if (instr->isVALU() || instr->isVINTRP())
       return true;
    if (!instr->isSALU() || instr->isSOPP())
       return false;
@@ -211,7 +213,7 @@ get_vopd_info(const SchedILPContext& ctx, const Instruction* instr)
 }
 
 bool
-is_vopd_compatible(const VOPDInfo& a, const VOPDInfo& b)
+is_vopd_compatible(const VOPDInfo& a, const VOPDInfo& b, bool* swap)
 {
    if ((a.is_opy_only && b.is_opy_only) || (a.is_dst_odd == b.is_dst_odd))
       return false;
@@ -220,6 +222,8 @@ is_vopd_compatible(const VOPDInfo& a, const VOPDInfo& b)
    if (a.has_literal && b.has_literal && a.literal != b.literal)
       return false;
 
+   *swap = false;
+
    /* The rest is checking src VGPR bank compatibility. */
    if ((a.src_banks & b.src_banks) == 0)
       return true;
@@ -242,11 +246,13 @@ is_vopd_compatible(const VOPDInfo& a, const VOPDInfo& b)
    if (b.op == aco_opcode::v_dual_mov_b32 && !a.is_commutative && a.is_opy_only)
       return false;
 
+   *swap = true;
+
    return true;
 }
 
 bool
-can_use_vopd(const SchedILPContext& ctx, unsigned idx)
+can_use_vopd(const SchedILPContext& ctx, unsigned idx, bool* prev_can_be_opx)
 {
    VOPDInfo cur_vopd = ctx.vopd[idx];
    Instruction* first = ctx.nodes[idx].instr;
@@ -258,9 +264,14 @@ can_use_vopd(const SchedILPContext& ctx, unsigned idx)
    if (ctx.prev_vopd_info.op == aco_opcode::num_opcodes || cur_vopd.op == aco_opcode::num_opcodes)
       return false;
 
-   if (!is_vopd_compatible(ctx.prev_vopd_info, cur_vopd))
+   bool swap = false;
+   if (!is_vopd_compatible(ctx.prev_vopd_info, cur_vopd, &swap))
       return false;
 
+   /* If we have to swap a v_mov_b32, it will become an OPY-only opcode. */
+   if (swap && !ctx.prev_vopd_info.is_commutative && cur_vopd.op == aco_opcode::v_dual_mov_b32)
+      cur_vopd.is_opy_only = true;
+
    assert(first->definitions.size() == 1);
    assert(first->definitions[0].size() == 1);
    assert(second->definitions.size() == 1);
@@ -277,26 +288,51 @@ can_use_vopd(const SchedILPContext& ctx, unsigned idx)
          return false;
    }
 
-   /* WaR dependencies are not a concern. */
-   return true;
+   /* WaR dependencies are not a concern before GFX12. */
+   *prev_can_be_opx = true;
+   if (ctx.program->gfx_level >= GFX12) {
+      /* From RDNA4 ISA doc:
+       * The OPX instruction must not overwrite sources of the OPY instruction".
+       */
+      bool war = false;
+      for (Operand op : first->operands) {
+         assert(op.size() == 1);
+         if (second->definitions[0].physReg() == op.physReg())
+            war = true;
+      }
+      if (war)
+         *prev_can_be_opx = false;
+   }
+
+   return *prev_can_be_opx || !cur_vopd.is_opy_only;
 }
 
-unsigned
-get_latency(const Instruction* const instr)
+Instruction_cycle_info
+get_cycle_info_with_mem_latency(const SchedILPContext& ctx, const Instruction* const instr)
 {
-   /* Note, that these are not accurate latency estimations. */
-   if (instr->isVALU())
-      return 5;
-   if (instr->isSALU())
-      return 2;
-   if (instr->isVMEM() || instr->isFlatLike())
-      return 32;
-   if (instr->isSMEM())
-      return 5;
-   if (instr->accessesLDS())
-      return 2;
+   Instruction_cycle_info cycle_info = get_cycle_info(*ctx.program, *instr);
 
-   return 0;
+   /* Based on get_wait_counter_info in aco_statistics.cpp. */
+   if (instr->isVMEM() || instr->isFlatLike()) {
+      cycle_info.latency = 320;
+   } else if (instr->isSMEM()) {
+      if (instr->operands.empty()) {
+         cycle_info.latency = 1;
+      } else if (instr->operands[0].size() == 2 ||
+                 (instr->operands[1].isConstant() &&
+                  (instr->operands.size() < 3 || instr->operands[2].isConstant()))) {
+         /* Likely cached. */
+         cycle_info.latency = 30;
+      } else {
+         cycle_info.latency = 200;
+      }
+   } else if (instr->isLDSDIR()) {
+      cycle_info.latency = 13;
+   } else if (instr->isDS()) {
+      cycle_info.latency = 20;
+   }
+
+   return cycle_info;
 }
 
 bool
@@ -305,7 +341,8 @@ is_memory_instr(const Instruction* const instr)
    /* For memory instructions, we allow to reorder them with ALU if it helps
     * to form larger clauses or to increase def-use distances.
     */
-   return instr->isVMEM() || instr->isFlatLike() || instr->isSMEM() || instr->accessesLDS();
+   return instr->isVMEM() || instr->isFlatLike() || instr->isSMEM() || instr->accessesLDS() ||
+          instr->isEXP();
 }
 
 constexpr unsigned max_sgpr = 128;
@@ -316,7 +353,8 @@ add_entry(SchedILPContext& ctx, Instruction* const instr, const uint32_t idx)
 {
    InstrInfo& entry = ctx.nodes[idx];
    entry.instr = instr;
-   entry.priority = 0;
+   entry.wait_cycles = 0;
+   entry.write_for_read_mask = 0;
    const mask_t mask = BITFIELD_BIT(idx);
    bool reorder = can_reorder(instr);
    ctx.active_mask |= mask;
@@ -345,28 +383,12 @@ add_entry(SchedILPContext& ctx, Instruction* const instr, const uint32_t idx)
          /* Add register reads. */
          reg_info.read_mask |= mask;
 
-         int cycles_since_reg_write = num_nodes;
          if (reg_info.has_direct_dependency) {
             /* A previous dependency is still part of the DAG. */
+            ctx.nodes[ctx.regs[reg].direct_dependency].write_for_read_mask |= mask;
             entry.dependency_mask |= BITFIELD_BIT(reg_info.direct_dependency);
-            cycles_since_reg_write = ctx.nodes[reg_info.direct_dependency].priority;
-         }
-
-         if (reg_info.latency) {
-            /* Ignore and reset register latencies for memory loads and other non-reorderable
-             * instructions. We schedule these as early as possible anyways.
-             */
-            if (reorder && reg_info.latency > cycles_since_reg_write) {
-               entry.priority = MIN2(entry.priority, cycles_since_reg_write - reg_info.latency);
-
-               /* If a previous register write created some latency, ensure that this
-                * is the first read of the register by making this instruction a direct
-                * dependency of all following register reads.
-                */
-               reg_info.has_direct_dependency = 1;
-               reg_info.direct_dependency = idx;
-            }
-            reg_info.latency = 0;
+         } else if (BITSET_TEST(ctx.reg_has_latency, reg + i)) {
+            entry.wait_cycles = MAX2(entry.wait_cycles, reg_info.latency);
          }
       }
    }
@@ -374,35 +396,35 @@ add_entry(SchedILPContext& ctx, Instruction* const instr, const uint32_t idx)
    /* Check if this instructions reads implicit registers. */
    if (needs_exec_mask(instr)) {
       for (unsigned reg = exec_lo; reg <= exec_hi; reg++) {
-         if (ctx.regs[reg].has_direct_dependency)
+         if (ctx.regs[reg].has_direct_dependency) {
             entry.dependency_mask |= BITFIELD_BIT(ctx.regs[reg].direct_dependency);
+            ctx.nodes[ctx.regs[reg].direct_dependency].write_for_read_mask |= mask;
+         }
          ctx.regs[reg].read_mask |= mask;
       }
    }
    if (ctx.program->gfx_level < GFX10 && instr->isScratch()) {
       for (unsigned reg = flat_scr_lo; reg <= flat_scr_hi; reg++) {
-         if (ctx.regs[reg].has_direct_dependency)
+         if (ctx.regs[reg].has_direct_dependency) {
             entry.dependency_mask |= BITFIELD_BIT(ctx.regs[reg].direct_dependency);
+            ctx.nodes[ctx.regs[reg].direct_dependency].write_for_read_mask |= mask;
+         }
          ctx.regs[reg].read_mask |= mask;
       }
    }
 
+   mask_t write_dep_mask = 0;
    for (const Definition& def : instr->definitions) {
       for (unsigned i = 0; i < def.size(); i++) {
          RegisterInfo& reg_info = ctx.regs[def.physReg().reg() + i];
 
          /* Add all previous register reads and writes to the dependencies. */
-         entry.dependency_mask |= reg_info.read_mask;
+         write_dep_mask |= reg_info.read_mask;
          reg_info.read_mask = mask;
 
          /* This register write is a direct dependency for all following reads. */
          reg_info.has_direct_dependency = 1;
          reg_info.direct_dependency = idx;
-
-         if (!ctx.is_vopd) {
-            /* Add latency information for the next register read. */
-            reg_info.latency = get_latency(instr);
-         }
       }
    }
 
@@ -422,19 +444,20 @@ add_entry(SchedILPContext& ctx, Instruction* const instr, const uint32_t idx)
       if (!is_memory_instr(instr) || instr->definitions.empty() ||
           get_sync_info(instr).semantics & semantic_volatile || ctx.is_vopd) {
          /* Add all previous instructions as dependencies. */
-         entry.dependency_mask = ctx.active_mask;
+         entry.dependency_mask = ctx.active_mask & ~ctx.non_reorder_mask;
       }
 
       /* Remove non-reorderable instructions from dependencies, since WaR dependencies can interfere
        * with clause formation. This should be fine, since these are always scheduled in-order and
        * any cases that are actually a concern for clause formation are added as transitive
        * dependencies. */
-      entry.dependency_mask &= ~ctx.non_reorder_mask;
-      entry.potential_clause = true;
+      write_dep_mask &= ~ctx.non_reorder_mask;
+      ctx.potential_partial_clause = true;
    } else if (ctx.last_non_reorderable != UINT8_MAX) {
-      ctx.nodes[ctx.last_non_reorderable].potential_clause = false;
+      ctx.potential_partial_clause = false;
    }
 
+   entry.dependency_mask |= write_dep_mask;
    entry.dependency_mask &= ~mask;
 
    for (unsigned i = 0; i < num_nodes; i++) {
@@ -444,9 +467,6 @@ add_entry(SchedILPContext& ctx, Instruction* const instr, const uint32_t idx)
       /* Add transitive dependencies. */
       if (entry.dependency_mask & BITFIELD_BIT(i))
          entry.dependency_mask |= ctx.nodes[i].dependency_mask;
-
-      /* increment base priority */
-      ctx.nodes[i].priority++;
    }
 }
 
@@ -456,6 +476,29 @@ remove_entry(SchedILPContext& ctx, const Instruction* const instr, const uint32_
    const mask_t mask = ~BITFIELD_BIT(idx);
    ctx.active_mask &= mask;
 
+   int latency = 0;
+   int stall = 1;
+   if (!ctx.is_vopd) {
+      Instruction_cycle_info cycle_info = get_cycle_info_with_mem_latency(ctx, instr);
+      latency = cycle_info.latency;
+      stall = cycle_info.issue_cycles;
+
+      if (ctx.nodes[idx].wait_cycles > 0) {
+         /* Add remaining latency stall. */
+         stall += ctx.nodes[idx].wait_cycles;
+      }
+
+      unsigned i;
+      BITSET_FOREACH_SET (i, ctx.reg_has_latency, 512) {
+         if (ctx.regs[i].latency <= stall) {
+            ctx.regs[i].latency = 0;
+            BITSET_CLEAR(ctx.reg_has_latency, i);
+         } else {
+            ctx.regs[i].latency -= stall;
+         }
+      }
+   }
+
    for (const Operand& op : instr->operands) {
       const unsigned reg = op.physReg();
       if (reg >= max_sgpr && reg != scc && reg < min_vgpr)
@@ -464,7 +507,6 @@ remove_entry(SchedILPContext& ctx, const Instruction* const instr, const uint32_
       for (unsigned i = 0; i < op.size(); i++) {
          RegisterInfo& reg_info = ctx.regs[reg + i];
          reg_info.read_mask &= mask;
-         reg_info.has_direct_dependency &= reg_info.direct_dependency != idx;
       }
    }
    if (needs_exec_mask(instr)) {
@@ -475,22 +517,36 @@ remove_entry(SchedILPContext& ctx, const Instruction* const instr, const uint32_
       ctx.regs[flat_scr_lo].read_mask &= mask;
       ctx.regs[flat_scr_hi].read_mask &= mask;
    }
+
    for (const Definition& def : instr->definitions) {
       for (unsigned i = 0; i < def.size(); i++) {
          unsigned reg = def.physReg().reg() + i;
          ctx.regs[reg].read_mask &= mask;
-         ctx.regs[reg].has_direct_dependency &= ctx.regs[reg].direct_dependency != idx;
+         if (ctx.regs[reg].has_direct_dependency && ctx.regs[reg].direct_dependency == idx) {
+            ctx.regs[reg].has_direct_dependency = false;
+            if (!ctx.is_vopd) {
+               BITSET_SET(ctx.reg_has_latency, reg);
+               ctx.regs[reg].latency = latency;
+            }
+         }
       }
    }
 
-   for (unsigned i = 0; i < num_nodes; i++)
+   for (unsigned i = 0; i < num_nodes; i++) {
       ctx.nodes[i].dependency_mask &= mask;
+      ctx.nodes[i].wait_cycles -= stall;
+      if (ctx.nodes[idx].write_for_read_mask & BITFIELD_BIT(i) && !ctx.is_vopd) {
+         ctx.nodes[i].wait_cycles = MAX2(ctx.nodes[i].wait_cycles, latency);
+      }
+   }
 
    if (ctx.next_non_reorderable == idx) {
       ctx.non_reorder_mask &= mask;
       ctx.next_non_reorderable = ctx.nodes[idx].next_non_reorderable;
-      if (ctx.last_non_reorderable == idx)
+      if (ctx.last_non_reorderable == idx) {
          ctx.last_non_reorderable = UINT8_MAX;
+         ctx.potential_partial_clause = false;
+      }
    }
 }
 
@@ -505,7 +561,11 @@ collect_clause_dependencies(const SchedILPContext& ctx, const uint8_t next, mask
 {
    const InstrInfo& entry = ctx.nodes[next];
    mask_t dependencies = entry.dependency_mask;
-   clause_mask |= (entry.potential_clause << next);
+   clause_mask |= BITFIELD_BIT(next);
+
+   /* If we dependent on the clause, don't add our dependencies. */
+   if (dependencies & clause_mask)
+      return 0;
 
    if (!is_memory_instr(entry.instr))
       return dependencies;
@@ -515,20 +575,13 @@ collect_clause_dependencies(const SchedILPContext& ctx, const uint8_t next, mask
     * instructions as dependencies. This prevents splitting of larger, already
     * formed clauses.
     */
-   if (next == ctx.last_non_reorderable && entry.potential_clause)
+   if (next == ctx.last_non_reorderable && ctx.potential_partial_clause)
       return (~clause_mask & ctx.active_mask) | dependencies;
 
-   if (entry.next_non_reorderable == UINT8_MAX)
-      return dependencies;
-
    /* Check if this can form a clause with the following non-reorderable instruction */
-   if (should_form_clause(entry.instr, ctx.nodes[entry.next_non_reorderable].instr)) {
-      mask_t clause_deps =
-         collect_clause_dependencies(ctx, entry.next_non_reorderable, clause_mask);
-
-      /* if the following clause is independent from us, add their dependencies */
-      if (!(clause_deps & BITFIELD_BIT(next)))
-         dependencies |= clause_deps;
+   if (entry.next_non_reorderable != UINT8_MAX &&
+       should_form_clause(entry.instr, ctx.nodes[entry.next_non_reorderable].instr)) {
+      dependencies |= collect_clause_dependencies(ctx, entry.next_non_reorderable, clause_mask);
    }
 
    return dependencies;
@@ -542,19 +595,26 @@ select_instruction_ilp(const SchedILPContext& ctx)
 {
    mask_t mask = ctx.active_mask;
 
-   /* First, collect all dependencies of the next non-reorderable instruction(s).
+   /* First, continue the currently open clause.
+    * Otherwise collect all dependencies of the next non-reorderable instruction(s).
     * These make up the list of possible candidates.
     */
-   if (ctx.next_non_reorderable != UINT8_MAX)
+   if (ctx.next_non_reorderable != UINT8_MAX) {
+      if (ctx.prev_info.instr && ctx.nodes[ctx.next_non_reorderable].dependency_mask == 0 &&
+          should_form_clause(ctx.prev_info.instr, ctx.nodes[ctx.next_non_reorderable].instr))
+         return ctx.next_non_reorderable;
       mask = collect_clause_dependencies(ctx, ctx.next_non_reorderable, 0);
+   }
 
-   /* If the next non-reorderable instruction has no dependencies, select it */
-   if (mask == 0)
-      return ctx.next_non_reorderable;
+   /* VINTRP(gfx6-10.3) can be handled like alu, but switching between VINTRP and other
+    * alu has a cost. So if the previous instr was VINTRP, try to keep selecting VINTRP.
+    */
+   bool prefer_vintrp = ctx.prev_info.instr && ctx.prev_info.instr->isVINTRP();
 
-   /* Otherwise, select the instruction with highest priority of all candidates. */
+   /* Select the instruction with lowest wait_cycles of all candidates. */
    unsigned idx = -1u;
-   int32_t priority = INT32_MIN;
+   bool idx_vintrp = false;
+   int32_t wait_cycles = INT32_MAX;
    u_foreach_bit (i, mask) {
       const InstrInfo& candidate = ctx.nodes[i];
 
@@ -562,21 +622,30 @@ select_instruction_ilp(const SchedILPContext& ctx)
       if (candidate.dependency_mask)
          continue;
 
-      if (idx == -1u || candidate.priority > priority) {
+      bool is_vintrp = prefer_vintrp && candidate.instr->isVINTRP();
+
+      if (idx == -1u || (is_vintrp && !idx_vintrp) ||
+          (is_vintrp == idx_vintrp && candidate.wait_cycles < wait_cycles)) {
          idx = i;
-         priority = candidate.priority;
+         idx_vintrp = is_vintrp;
+         wait_cycles = candidate.wait_cycles;
       }
    }
 
-   assert(idx != -1u);
-   return idx;
+   if (idx != -1u)
+      return idx;
+
+   /* Select the next non-reorderable instruction. (it must have no dependencies) */
+   assert(ctx.next_non_reorderable != UINT8_MAX);
+   assert(ctx.nodes[ctx.next_non_reorderable].dependency_mask == 0);
+   return ctx.next_non_reorderable;
 }
 
 bool
 compare_nodes_vopd(const SchedILPContext& ctx, int num_vopd_odd_minus_even, bool* use_vopd,
-                   unsigned current, unsigned candidate)
+                   bool* prev_can_be_opx, unsigned current, unsigned candidate)
 {
-   if (can_use_vopd(ctx, candidate)) {
+   if (can_use_vopd(ctx, candidate, prev_can_be_opx)) {
       /* If we can form a VOPD instruction, always prefer to do so. */
       if (!*use_vopd) {
          *use_vopd = true;
@@ -608,11 +677,11 @@ compare_nodes_vopd(const SchedILPContext& ctx, int num_vopd_odd_minus_even, bool
       }
    }
 
-   return ctx.nodes[candidate].priority > ctx.nodes[current].priority;
+   return ctx.nodes[candidate].wait_cycles < ctx.nodes[current].wait_cycles;
 }
 
 unsigned
-select_instruction_vopd(const SchedILPContext& ctx, bool* use_vopd)
+select_instruction_vopd(const SchedILPContext& ctx, bool* use_vopd, bool* prev_can_be_opx)
 {
    *use_vopd = false;
 
@@ -634,11 +703,14 @@ select_instruction_vopd(const SchedILPContext& ctx, bool* use_vopd)
       if (candidate.dependency_mask)
          continue;
 
+      bool prev_can_be_opx_for_i;
       if (cur == -1u) {
          cur = i;
-         *use_vopd = can_use_vopd(ctx, i);
-      } else if (compare_nodes_vopd(ctx, num_vopd_odd_minus_even, use_vopd, cur, i)) {
+         *use_vopd = can_use_vopd(ctx, i, prev_can_be_opx);
+      } else if (compare_nodes_vopd(ctx, num_vopd_odd_minus_even, use_vopd, &prev_can_be_opx_for_i,
+                                    cur, i)) {
          cur = i;
+         *prev_can_be_opx = prev_can_be_opx_for_i;
       }
    }
 
@@ -674,24 +746,29 @@ get_vopd_opcode_operands(const SchedILPContext& ctx, Instruction* instr, const V
 }
 
 Instruction*
-create_vopd_instruction(const SchedILPContext& ctx, unsigned idx)
+create_vopd_instruction(const SchedILPContext& ctx, unsigned idx, bool prev_can_be_opx)
 {
    Instruction* x = ctx.prev_info.instr;
    Instruction* y = ctx.nodes[idx].instr;
    VOPDInfo x_info = ctx.prev_vopd_info;
    VOPDInfo y_info = ctx.vopd[idx];
+   x_info.is_opy_only |= !prev_can_be_opx;
 
    bool swap_x = false, swap_y = false;
    if (x_info.src_banks & y_info.src_banks) {
       assert(x_info.is_commutative || y_info.is_commutative);
       /* Avoid swapping v_mov_b32 because it will become an OPY-only opcode. */
-      if (x_info.op == aco_opcode::v_dual_mov_b32 && !y_info.is_commutative) {
+      if (x_info.op == aco_opcode::v_dual_mov_b32 && y_info.op == aco_opcode::v_dual_mov_b32) {
+         swap_x = x_info.is_opy_only;
+         swap_y = !swap_x;
+      } else if (x_info.op == aco_opcode::v_dual_mov_b32 && !y_info.is_commutative) {
          swap_x = true;
          x_info.is_opy_only = true;
       } else {
          swap_x = x_info.is_commutative && x_info.op != aco_opcode::v_dual_mov_b32;
          swap_y = y_info.is_commutative && !swap_x;
       }
+      y_info.is_opy_only |= swap_y && y_info.op == aco_opcode::v_dual_mov_b32;
    }
 
    if (x_info.is_opy_only) {
@@ -699,6 +776,7 @@ create_vopd_instruction(const SchedILPContext& ctx, unsigned idx)
       std::swap(x_info, y_info);
       std::swap(swap_x, swap_y);
    }
+   assert(!x_info.is_opy_only);
 
    aco_opcode x_op, y_op;
    unsigned num_operands = 0;
@@ -729,14 +807,15 @@ do_schedule(SchedILPContext& ctx, It& insert_it, It& remove_it, It instructions_
 
    ctx.prev_info.instr = NULL;
    bool use_vopd = false;
+   bool prev_can_be_opx;
 
    while (ctx.active_mask) {
-      unsigned next_idx =
-         ctx.is_vopd ? select_instruction_vopd(ctx, &use_vopd) : select_instruction_ilp(ctx);
+      unsigned next_idx = ctx.is_vopd ? select_instruction_vopd(ctx, &use_vopd, &prev_can_be_opx)
+                                      : select_instruction_ilp(ctx);
       Instruction* next_instr = ctx.nodes[next_idx].instr;
 
       if (use_vopd) {
-         std::prev(insert_it)->reset(create_vopd_instruction(ctx, next_idx));
+         std::prev(insert_it)->reset(create_vopd_instruction(ctx, next_idx, prev_can_be_opx));
          ctx.prev_info.instr = NULL;
       } else {
          (insert_it++)->reset(next_instr);
@@ -750,7 +829,7 @@ do_schedule(SchedILPContext& ctx, It& insert_it, It& remove_it, It instructions_
       if (remove_it != instructions_end) {
          add_entry(ctx, (remove_it++)->get(), next_idx);
       } else if (ctx.last_non_reorderable != UINT8_MAX) {
-         ctx.nodes[ctx.last_non_reorderable].potential_clause = false;
+         ctx.potential_partial_clause = false;
          ctx.last_non_reorderable = UINT8_MAX;
       }
    }
@@ -764,10 +843,14 @@ schedule_ilp(Program* program)
    SchedILPContext ctx = {program};
 
    for (Block& block : program->blocks) {
+      if (block.instructions.empty())
+         continue;
       auto it = block.instructions.begin();
       auto insert_it = block.instructions.begin();
       do_schedule(ctx, insert_it, it, block.instructions.begin(), block.instructions.end());
       block.instructions.resize(insert_it - block.instructions.begin());
+      if (block.linear_succs.empty() || block.instructions.back()->opcode == aco_opcode::s_branch)
+         BITSET_ZERO(ctx.reg_has_latency);
    }
 }
 
