@@ -66,7 +66,7 @@ bool
 setup_nir_cs(enum amd_gfx_level gfx_level, gl_shader_stage stage = MESA_SHADER_COMPUTE,
              enum radeon_family family = CHIP_UNKNOWN, const char* subvariant = "");
 
-void finish_program(aco::Program* program, bool endpgm = true);
+void finish_program(aco::Program* program, bool endpgm = true, bool dominance = false);
 void finish_validator_test();
 void finish_opt_test();
 void finish_setup_reduce_temp_test();
@@ -98,7 +98,9 @@ aco::Temp fsat(aco::Temp src, aco::Builder b = bld);
 aco::Temp fmin(aco::Temp src0, aco::Temp src1, aco::Builder b = bld);
 aco::Temp fmax(aco::Temp src0, aco::Temp src1, aco::Builder b = bld);
 aco::Temp ext_ushort(aco::Temp src, unsigned idx, aco::Builder b = bld);
+aco::Temp ext_sshort(aco::Temp src, unsigned idx, aco::Builder b = bld);
 aco::Temp ext_ubyte(aco::Temp src, unsigned idx, aco::Builder b = bld);
+aco::Temp ext_sbyte(aco::Temp src, unsigned idx, aco::Builder b = bld);
 void emit_divergent_if_else(aco::Program* prog, aco::Builder& b, aco::Operand cond,
                             std::function<void()> then, std::function<void()> els);
 
