@@ -10,3 +10,8 @@ DRI_CONF_SECTION_PERFORMANCE
     */
    DRI_CONF_OPT_I(pan_csf_max_chunks, 64, 1, 65535, "CSF Tiler Max Chunks")
 DRI_CONF_SECTION_END
+
+DRI_CONF_SECTION_MISCELLANEOUS
+   DRI_CONF_PAN_COMPUTE_CORE_MASK(~0ull)
+   DRI_CONF_PAN_FRAGMENT_CORE_MASK(~0ull)
+DRI_CONF_SECTION_END
