@@ -146,12 +146,22 @@ nouveau_zink_predicate(int fd, const char *driver)
 
    bool prefer_zink = false;
 
-   /* enable this once zink is up to speed.
-    * struct drm_nouveau_getparam r = { .param = NOUVEAU_GETPARAM_CHIPSET_ID };
-    * int ret = drmCommandWriteRead(fd, DRM_NOUVEAU_GETPARAM, &r, sizeof(r));
-    * if (ret == 0 && (r.value & ~0xf) >= 0x160)
-    *    prefer_zink = true;
+   /* Enable Zink by default on Turing and later GPUs
+    *
+    * We only use Zink if if the kernel supports VMA_TILEMODE, which is needed
+    * for DRM format modifiers.  This also doubles as a check for a new enough
+    * kernel to run NVK in general.
     */
+   struct drm_nouveau_getparam r = { .param = NOUVEAU_GETPARAM_HAS_VMA_TILEMODE };
+   int ret = drmCommandWriteRead(fd, DRM_NOUVEAU_GETPARAM, &r, sizeof(r));
+   if (ret == 0 && r.value == 1) {
+      r.param = NOUVEAU_GETPARAM_CHIPSET_ID;
+      r.value = 0;
+      ret = drmCommandWriteRead(fd, DRM_NOUVEAU_GETPARAM, &r, sizeof(r));
+      if (ret == 0 && r.value >= 0x160) {
+         prefer_zink = true;
+      }
+   }
 
    prefer_zink = debug_get_bool_option("NOUVEAU_USE_ZINK", prefer_zink);
 
@@ -173,16 +183,46 @@ nouveau_zink_predicate(int fd, const char *driver)
 int
 loader_open_render_node_platform_device(const char * const drivers[],
                                         unsigned int n_drivers)
+{
+   unsigned int n_devices;
+   int *fds = loader_open_render_node_platform_devices(drivers, n_drivers, &n_devices);
+   int fd = -1;
+
+   if (n_devices > 0) {
+      fd = fds[0];
+      free(fds);
+   }
+
+   return fd;
+}
+
+/**
+ * Goes through all the platform devices whose driver is on the given list and
+ * try to open their render node. It returns an array with the fds of all the
+ * devices that it can open.
+ *
+ * Caller must close the returned fds and free the array.
+ */
+int *
+loader_open_render_node_platform_devices(const char * const drivers[],
+                                         unsigned int n_drivers,
+                                         unsigned int *n_devices)
 {
    drmDevicePtr devices[MAX_DRM_DEVICES], device;
    int num_devices, fd = -1;
    int i, j;
    bool found = false;
+   int *result;
 
    num_devices = drmGetDevices2(0, devices, MAX_DRM_DEVICES);
-   if (num_devices <= 0)
-      return -ENOENT;
+   if (num_devices <= 0) {
+      *n_devices = 0;
+      return NULL;
+   }
 
+   result = calloc(n_drivers, num_devices);
+
+   *n_devices = 0;
    for (i = 0; i < num_devices; i++) {
       device = devices[i];
 
@@ -206,22 +246,23 @@ loader_open_render_node_platform_device(const char * const drivers[],
                break;
             }
          }
-         if (!found) {
-            drmFreeVersion(version);
-            close(fd);
-            continue;
-         }
 
          drmFreeVersion(version);
-         break;
+
+         if (found)
+            result[(*n_devices)++] = fd;
+         else
+            close(fd);
       }
    }
    drmFreeDevices(devices, num_devices);
 
-   if (i == num_devices)
-      return -ENOENT;
+   if (*n_devices == 0) {
+      free(result);
+      return NULL;
+   }
 
-   return fd;
+   return result;
 }
 
 bool
@@ -815,18 +856,13 @@ loader_open_driver_lib(const char *driver_name,
          next = end;
 
       len = next - p;
-      snprintf(path, sizeof(path), "%.*s/tls/%s%s.so", len,
+      snprintf(path, sizeof(path), "%.*s/%s%s.so", len,
                p, driver_name, lib_suffix);
       driver = dlopen(path, RTLD_NOW | RTLD_GLOBAL);
       if (driver == NULL) {
-         snprintf(path, sizeof(path), "%.*s/%s%s.so", len,
-                  p, driver_name, lib_suffix);
-         driver = dlopen(path, RTLD_NOW | RTLD_GLOBAL);
-         if (driver == NULL) {
-            dl_error = dlerror();
-            log_(_LOADER_DEBUG, "MESA-LOADER: failed to open %s: %s\n",
-                 path, dl_error);
-         }
+         dl_error = dlerror();
+         log_(_LOADER_DEBUG, "MESA-LOADER: failed to open %s: %s\n",
+              path, dl_error);
       }
       /* not need continue to loop all paths once the driver is found */
       if (driver != NULL)
