@@ -83,8 +83,7 @@ static nir_def *
 get_layer_id(nir_builder *b)
 {
 #if PAN_ARCH <= 7
-   return nir_load_push_constant(b, 1, 32, nir_imm_int(b, 0), .base = 0,
-                                 .range = 4);
+   return nir_load_push_constant(b, 1, 32, nir_imm_int(b, 0));
 #else
    return nir_load_layer_id(b);
 #endif
@@ -172,7 +171,6 @@ get_preload_shader(struct panvk_device *dev,
 
    struct panfrost_compile_inputs inputs = {
       .gpu_id = phys_dev->kmod.props.gpu_prod_id,
-      .no_ubo_to_push = true,
       .is_blit = true,
    };
 
@@ -180,6 +178,8 @@ get_preload_shader(struct panvk_device *dev,
 
    VkResult result = panvk_per_arch(create_internal_shader)(
       dev, nir, &inputs, &shader);
+   ralloc_free(nir);
+
    if (result != VK_SUCCESS)
       return result;
 
@@ -190,7 +190,8 @@ get_preload_shader(struct panvk_device *dev,
       return panvk_error(dev, VK_ERROR_OUT_OF_DEVICE_MEMORY);
    }
 
-   pan_pack(panvk_priv_mem_host_addr(shader->spd), SHADER_PROGRAM, cfg) {
+   pan_cast_and_pack(panvk_priv_mem_host_addr(shader->spd), SHADER_PROGRAM,
+                     cfg) {
       cfg.stage = MALI_SHADER_STAGE_FRAGMENT;
       cfg.fragment_coverage_bitmask_type = MALI_COVERAGE_BITMASK_TYPE_GL;
       cfg.register_allocation = MALI_SHADER_REGISTER_ALLOCATION_32_PER_THREAD;
@@ -210,10 +211,8 @@ out:
 }
 
 static VkResult
-alloc_pre_post_dcds(struct panvk_cmd_buffer *cmdbuf)
+alloc_pre_post_dcds(struct panvk_cmd_buffer *cmdbuf, struct pan_fb_info *fbinfo)
 {
-   struct pan_fb_info *fbinfo = &cmdbuf->state.gfx.render.fb.info;
-
    if (fbinfo->bifrost.pre_post.dcds.gpu)
       return VK_SUCCESS;
 
@@ -244,19 +243,17 @@ get_reg_fmt(nir_alu_type type)
 }
 
 static void
-fill_textures(struct panvk_cmd_buffer *cmdbuf,
+fill_textures(struct panvk_cmd_buffer *cmdbuf, struct pan_fb_info *fbinfo,
               const struct panvk_fb_preload_shader_key *key,
               struct mali_texture_packed *textures)
 {
-   struct pan_fb_info *fbinfo = &cmdbuf->state.gfx.render.fb.info;
-
    if (key->aspects == VK_IMAGE_ASPECT_COLOR_BIT) {
       for (unsigned i = 0; i < fbinfo->rt_count; i++) {
          struct panvk_image_view *iview =
             cmdbuf->state.gfx.render.color_attachments.iviews[i];
 
          if (iview)
-            textures[i] = iview->descs.tex;
+            textures[i] = iview->descs.tex[0];
          else
             textures[i] = (struct mali_texture_packed){0};
       }
@@ -270,8 +267,8 @@ fill_textures(struct panvk_cmd_buffer *cmdbuf,
             ?: cmdbuf->state.gfx.render.s_attachment.iview;
 
       textures[idx++] = vk_format_has_depth(iview->vk.view_format)
-                           ? iview->descs.tex
-                           : iview->descs.other_aspect_tex;
+                           ? iview->descs.zs.tex
+                           : iview->descs.zs.other_aspect_tex;
    }
 
    if (key->aspects & VK_IMAGE_ASPECT_STENCIL_BIT) {
@@ -280,17 +277,16 @@ fill_textures(struct panvk_cmd_buffer *cmdbuf,
             ?: cmdbuf->state.gfx.render.z_attachment.iview;
 
       textures[idx++] = vk_format_has_depth(iview->vk.view_format)
-                           ? iview->descs.other_aspect_tex
-                           : iview->descs.tex;
+                           ? iview->descs.zs.other_aspect_tex
+                           : iview->descs.zs.tex;
    }
 }
 
 static void
-fill_bds(struct panvk_cmd_buffer *cmdbuf,
+fill_bds(struct pan_fb_info *fbinfo,
          const struct panvk_fb_preload_shader_key *key,
          struct mali_blend_packed *bds)
 {
-   struct pan_fb_info *fbinfo = &cmdbuf->state.gfx.render.fb.info;
    uint32_t bd_count = MAX2(fbinfo->rt_count, 1);
 
    for (unsigned i = 0; i < bd_count; i++) {
@@ -329,11 +325,10 @@ fill_bds(struct panvk_cmd_buffer *cmdbuf,
 
 #if PAN_ARCH <= 7
 static VkResult
-cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
+cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf, struct pan_fb_info *fbinfo,
              const struct panvk_fb_preload_shader_key *key)
 {
    struct panvk_device *dev = to_panvk_device(cmdbuf->vk.base.device);
-   struct pan_fb_info *fbinfo = &cmdbuf->state.gfx.render.fb.info;
    struct panvk_internal_shader *shader = NULL;
 
    VkResult result = get_preload_shader(dev, key, &shader);
@@ -351,7 +346,7 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
    if (!rsd.cpu)
       return VK_ERROR_OUT_OF_DEVICE_MEMORY;
 
-   pan_pack(rsd.cpu, RENDERER_STATE, cfg) {
+   pan_cast_and_pack(rsd.cpu, RENDERER_STATE, cfg) {
       pan_shader_prepare_rsd(&shader->info,
                              panvk_priv_mem_dev_addr(shader->code_mem), &cfg);
 
@@ -380,7 +375,7 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
 
       if (key->aspects == VK_IMAGE_ASPECT_COLOR_BIT) {
          /* Skipping ATEST requires forcing Z/S */
-         cfg.properties.zs_update_operation = MALI_PIXEL_KILL_STRONG_EARLY;
+         cfg.properties.zs_update_operation = MALI_PIXEL_KILL_FORCE_EARLY;
          cfg.properties.pixel_kill_operation = MALI_PIXEL_KILL_FORCE_EARLY;
       } else {
          /* Writing Z/S requires late updates */
@@ -399,7 +394,7 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
             key->aspects == VK_IMAGE_ASPECT_COLOR_BIT;
    }
 
-   fill_bds(cmdbuf, key, rsd.cpu + pan_size(RENDERER_STATE));
+   fill_bds(fbinfo, key, rsd.cpu + pan_size(RENDERER_STATE));
 
    struct panvk_batch *batch = cmdbuf->cur_batch;
    uint16_t minx = 0, miny = 0, maxx, maxy;
@@ -414,7 +409,7 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
    if (!vpd.cpu)
       return VK_ERROR_OUT_OF_DEVICE_MEMORY;
 
-   pan_pack(vpd.cpu, VIEWPORT, cfg) {
+   pan_cast_and_pack(vpd.cpu, VIEWPORT, cfg) {
       cfg.scissor_minimum_x = minx;
       cfg.scissor_minimum_y = miny;
       cfg.scissor_maximum_x = maxx;
@@ -425,9 +420,10 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
    if (!sampler.cpu)
       return VK_ERROR_OUT_OF_DEVICE_MEMORY;
 
-   pan_pack(sampler.cpu, SAMPLER, cfg) {
+   pan_cast_and_pack(sampler.cpu, SAMPLER, cfg) {
       cfg.seamless_cube_map = false;
       cfg.normalized_coordinates = false;
+      cfg.clamp_integer_array_indices = false;
       cfg.minify_nearest = true;
       cfg.magnify_nearest = true;
    }
@@ -437,9 +433,9 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
    if (!textures.cpu)
       return VK_ERROR_OUT_OF_DEVICE_MEMORY;
 
-   fill_textures(cmdbuf, key, textures.cpu);
+   fill_textures(cmdbuf, fbinfo, key, textures.cpu);
 
-   result = alloc_pre_post_dcds(cmdbuf);
+   result = alloc_pre_post_dcds(cmdbuf, fbinfo);
    if (result != VK_SUCCESS)
       return result;
 
@@ -492,10 +488,18 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
       fbinfo->bifrost.pre_post.modes[dcd_idx] =
          MALI_PRE_POST_FRAME_SHADER_MODE_INTERSECT;
    } else {
-      enum pipe_format fmt = fbinfo->zs.view.zs
-                                ? fbinfo->zs.view.zs->planes[0]->layout.format
-                                : fbinfo->zs.view.s->planes[0]->layout.format;
-      bool always = false;
+      const struct pan_image *plane =
+         fbinfo->zs.view.zs ? pan_image_view_get_zs_plane(fbinfo->zs.view.zs)
+                            : pan_image_view_get_s_plane(fbinfo->zs.view.s);
+      enum pipe_format fmt = plane->layout.format;
+      /* On some GPUs (e.g. G31), we must use SHADER_MODE_ALWAYS rather than
+       * SHADER_MODE_INTERSECT for full screen operations. Since the full
+       * screen rectangle will always intersect, this won't affect
+       * performance.
+       */
+      bool always = !fbinfo->extent.minx && !fbinfo->extent.miny &&
+                    fbinfo->extent.maxx == (fbinfo->width - 1) &&
+                    fbinfo->extent.maxy == (fbinfo->height - 1);
 
       /* If we're dealing with a combined ZS resource and only one
        * component is cleared, we need to reload the whole surface
@@ -513,11 +517,21 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
        * Thing's haven't been benchmarked to determine what's
        * preferable (saving bandwidth vs having ZS preloaded
        * earlier), so let's leave it like that for now.
+       * HOWEVER, EARLY_ZS_ALWAYS doesn't exist on 7.0, only on
+       * 7.2 and later, so check for that!
        */
-      fbinfo->bifrost.pre_post.modes[dcd_idx] =
-         PAN_ARCH > 6
-            ? MALI_PRE_POST_FRAME_SHADER_MODE_EARLY_ZS_ALWAYS
-         : always ? MALI_PRE_POST_FRAME_SHADER_MODE_ALWAYS
+      struct panvk_physical_device *pdev =
+         to_panvk_physical_device(dev->vk.physical);
+      unsigned gpu_id = pdev->kmod.props.gpu_prod_id;
+
+      /* the PAN_ARCH check is redundant but allows compiler optimization
+         when PAN_ARCH <= 6 */
+      if (PAN_ARCH > 6 && gpu_id >= 0x7200)
+         fbinfo->bifrost.pre_post.modes[dcd_idx] =
+            MALI_PRE_POST_FRAME_SHADER_MODE_EARLY_ZS_ALWAYS;
+      else
+         fbinfo->bifrost.pre_post.modes[dcd_idx] = always
+                  ? MALI_PRE_POST_FRAME_SHADER_MODE_ALWAYS
                   : MALI_PRE_POST_FRAME_SHADER_MODE_INTERSECT;
    }
 
@@ -525,11 +539,10 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
 }
 #else
 static VkResult
-cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
+cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf, struct pan_fb_info *fbinfo,
              struct panvk_fb_preload_shader_key *key)
 {
    struct panvk_device *dev = to_panvk_device(cmdbuf->vk.base.device);
-   struct pan_fb_info *fbinfo = &cmdbuf->state.gfx.render.fb.info;
    struct panvk_internal_shader *shader = NULL;
 
    VkResult result = get_preload_shader(dev, key, &shader);
@@ -558,20 +571,29 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
    pan_pack(sampler, SAMPLER, cfg) {
       cfg.seamless_cube_map = false;
       cfg.normalized_coordinates = false;
+      cfg.clamp_integer_array_indices = false;
       cfg.minify_nearest = true;
       cfg.magnify_nearest = true;
    }
 
-   fill_textures(cmdbuf, key, descs.cpu + PANVK_DESCRIPTOR_SIZE);
+   fill_textures(cmdbuf, fbinfo, key, descs.cpu + PANVK_DESCRIPTOR_SIZE);
+
+   uint32_t rt_written = 0;
+   if (key->aspects == VK_IMAGE_ASPECT_COLOR_BIT) {
+      for (unsigned i = 0; i < fbinfo->rt_count; i++) {
+         if (fbinfo->rts[i].preload)
+            rt_written |= BITFIELD_BIT(i);
+      }
+   }
 
    if (key->aspects == VK_IMAGE_ASPECT_COLOR_BIT)
-      fill_bds(cmdbuf, key, bds.cpu);
+      fill_bds(fbinfo, key, bds.cpu);
 
    struct panfrost_ptr res_table = panvk_cmd_alloc_desc(cmdbuf, RESOURCE);
    if (!res_table.cpu)
       return VK_ERROR_OUT_OF_DEVICE_MEMORY;
 
-   pan_pack(res_table.cpu, RESOURCE, cfg) {
+   pan_cast_and_pack(res_table.cpu, RESOURCE, cfg) {
       cfg.address = descs.gpu;
       cfg.size = desc_count * PANVK_DESCRIPTOR_SIZE;
    }
@@ -585,7 +607,7 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
    bool preload_s =
       key->aspects != VK_IMAGE_ASPECT_COLOR_BIT && fbinfo->zs.preload.s;
 
-   pan_pack(zsd.cpu, DEPTH_STENCIL, cfg) {
+   pan_cast_and_pack(zsd.cpu, DEPTH_STENCIL, cfg) {
       cfg.depth_function = MALI_FUNC_ALWAYS;
       cfg.depth_write_enable = preload_z;
 
@@ -612,7 +634,7 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
       cfg.depth_cull_enable = false;
    }
 
-   result = alloc_pre_post_dcds(cmdbuf);
+   result = alloc_pre_post_dcds(cmdbuf, fbinfo);
    if (result != VK_SUCCESS)
       return result;
 
@@ -622,32 +644,41 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
    pan_pack(&dcds[dcd_idx], DRAW, cfg) {
       if (key->aspects == VK_IMAGE_ASPECT_COLOR_BIT) {
          /* Skipping ATEST requires forcing Z/S */
-         cfg.zs_update_operation = MALI_PIXEL_KILL_STRONG_EARLY;
-         cfg.pixel_kill_operation = MALI_PIXEL_KILL_FORCE_EARLY;
+         cfg.flags_0.zs_update_operation = MALI_PIXEL_KILL_FORCE_EARLY;
+         cfg.flags_0.pixel_kill_operation = MALI_PIXEL_KILL_FORCE_EARLY;
 
          cfg.blend = bds.gpu;
          cfg.blend_count = bd_count;
-         cfg.render_target_mask = cmdbuf->state.gfx.render.bound_attachments &
-                                  MESA_VK_RP_ATTACHMENT_ANY_COLOR_BITS;
+         cfg.flags_1.render_target_mask =
+            cmdbuf->state.gfx.render.bound_attachments &
+            MESA_VK_RP_ATTACHMENT_ANY_COLOR_BITS;
       } else {
          /* ZS_EMIT requires late update/kill */
-         cfg.zs_update_operation = MALI_PIXEL_KILL_FORCE_LATE;
-         cfg.pixel_kill_operation = MALI_PIXEL_KILL_FORCE_LATE;
+         cfg.flags_0.zs_update_operation = MALI_PIXEL_KILL_FORCE_LATE;
+         cfg.flags_0.pixel_kill_operation = MALI_PIXEL_KILL_FORCE_LATE;
          cfg.blend_count = 0;
       }
 
-      cfg.allow_forward_pixel_to_kill =
+      cfg.flags_0.allow_forward_pixel_to_kill =
          key->aspects == VK_IMAGE_ASPECT_COLOR_BIT;
-      cfg.allow_forward_pixel_to_be_killed = true;
+      cfg.flags_0.allow_forward_pixel_to_be_killed = true;
       cfg.depth_stencil = zsd.gpu;
-      cfg.sample_mask = 0xFFFF;
-      cfg.multisample_enable = key->samples > 1;
-      cfg.evaluate_per_sample = key->samples > 1;
+      cfg.flags_1.sample_mask = 0xFFFF;
+      cfg.flags_0.multisample_enable = key->samples > 1;
+      cfg.flags_0.evaluate_per_sample = key->samples > 1;
+      cfg.flags_0.clean_fragment_write = true;
+
+#if PAN_ARCH >= 12
+      cfg.fragment_resources = res_table.gpu | 1;
+      cfg.fragment_shader = panvk_priv_mem_dev_addr(shader->spd);
+      cfg.thread_storage = cmdbuf->state.gfx.tsd;
+#else
       cfg.maximum_z = 1.0;
-      cfg.clean_fragment_write = true;
       cfg.shader.resources = res_table.gpu | 1;
       cfg.shader.shader = panvk_priv_mem_dev_addr(shader->spd);
       cfg.shader.thread_storage = cmdbuf->state.gfx.tsd;
+#endif
+      cfg.flags_2.write_mask = rt_written;
    }
 
    if (key->aspects == VK_IMAGE_ASPECT_COLOR_BIT) {
@@ -661,9 +692,17 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
        * Thing's haven't been benchmarked to determine what's
        * preferable (saving bandwidth vs having ZS preloaded
        * earlier), so let's leave it like that for now.
+       *
+       * On v13+, we don't have EARLY_ZS_ALWAYS instead we use
+       * PREPASS_ALWAYS.
        */
+#if PAN_ARCH >= 13
+      fbinfo->bifrost.pre_post.modes[dcd_idx] =
+         MALI_PRE_POST_FRAME_SHADER_MODE_PREPASS_ALWAYS;
+#else
       fbinfo->bifrost.pre_post.modes[dcd_idx] =
          MALI_PRE_POST_FRAME_SHADER_MODE_EARLY_ZS_ALWAYS;
+#endif
    }
 
    return VK_SUCCESS;
@@ -671,10 +710,9 @@ cmd_emit_dcd(struct panvk_cmd_buffer *cmdbuf,
 #endif
 
 static VkResult
-cmd_preload_zs_attachments(struct panvk_cmd_buffer *cmdbuf)
+cmd_preload_zs_attachments(struct panvk_cmd_buffer *cmdbuf,
+                           struct pan_fb_info *fbinfo)
 {
-   struct pan_fb_info *fbinfo = &cmdbuf->state.gfx.render.fb.info;
-
    if (!fbinfo->zs.preload.s && !fbinfo->zs.preload.z)
       return VK_SUCCESS;
 
@@ -705,13 +743,13 @@ cmd_preload_zs_attachments(struct panvk_cmd_buffer *cmdbuf)
       assert(key.view_type == view_type);
    }
 
-   return cmd_emit_dcd(cmdbuf, &key);
+   return cmd_emit_dcd(cmdbuf, fbinfo, &key);
 }
 
 static VkResult
-cmd_preload_color_attachments(struct panvk_cmd_buffer *cmdbuf)
+cmd_preload_color_attachments(struct panvk_cmd_buffer *cmdbuf,
+                              struct pan_fb_info *fbinfo)
 {
-   struct pan_fb_info *fbinfo = &cmdbuf->state.gfx.render.fb.info;
    struct panvk_fb_preload_shader_key key = {
       .type = PANVK_META_OBJECT_KEY_FB_PRELOAD_SHADER,
       .samples = fbinfo->nr_samples,
@@ -743,15 +781,16 @@ cmd_preload_color_attachments(struct panvk_cmd_buffer *cmdbuf)
    if (!needs_preload)
       return VK_SUCCESS;
 
-   return cmd_emit_dcd(cmdbuf, &key);
+   return cmd_emit_dcd(cmdbuf, fbinfo, &key);
 }
 
 VkResult
-panvk_per_arch(cmd_fb_preload)(struct panvk_cmd_buffer *cmdbuf)
+panvk_per_arch(cmd_fb_preload)(struct panvk_cmd_buffer *cmdbuf,
+                               struct pan_fb_info *fbinfo)
 {
-   VkResult result = cmd_preload_color_attachments(cmdbuf);
+   VkResult result = cmd_preload_color_attachments(cmdbuf, fbinfo);
    if (result != VK_SUCCESS)
       return result;
 
-   return cmd_preload_zs_attachments(cmdbuf);
+   return cmd_preload_zs_attachments(cmdbuf, fbinfo);
 }
